// -*- C++ -*-
/*!
 * @file ConfigAdmin.h
 * @brief Configuration Administration classes
 * @date $Date: 2007-12-31 03:08:02 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2007-2011
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: ConfigAdmin.h 2237 2011-10-25 01:25:37Z n-ando $
 *
 */

#ifndef RTC_CONFIGADMIN_H
#define RTC_CONFIGADMIN_H

#include <string>
#include <vector>
#include <iostream>
#include <coil/Properties.h>
#include <coil/stringutil.h>
#include <rtm/ConfigurationListener.h>

/*!
 * @if jp
 * @namespace RTC
 *
 * @brief RTR|[lg
 *
 * @else
 *
 * @namespace RTC
 *
 * @brief RT-Component
 *
 * @endif
 */
namespace RTC
{
  /*!
   * @if jp
   * @class OnUpdateCallback
   * @brief OnUpdate R[obNۃNX
   *
   * @else
   * @class OnUpdateCallback
   * @brief Callback functor abstract for OnUpdate
   *
   * @endif
   */
  typedef ConfigurationSetNameListener OnUpdateCallback;

  /*!
   * @if jp
   * @class OnUpdateParamCallback
   * @brief OnUpdateParam R[obNۃNX
   *
   * @else
   * @class OnUpdateParamCallback
   * @brief Callback functor abstract for OnUpdateParam
   *
   * @endif
   */
  typedef ConfigurationParamListener OnUpdateParamCallback;

  /*!
   * @if jp
   * @class OnSetConfigurationSetCallback
   * @brief OnSetConfigurationSet R[obNۃNX
   *
   * @else
   * @class OnSetConfigurationSetCallback
   * @brief Callback functor abstract for OnSetConfigurationSet
   *
   * @endif
   */
  typedef ConfigurationSetListener OnSetConfigurationSetCallback;

  /*!
   * @if jp
   * @class OnAddConfigurationAddCallback
   * @brief OnAddConfigurationAdd R[obNۃNX
   *
   * @else
   * @class OnAddConfigurationAddCallback
   * @brief callback functor abstract for OnAddConfigurationAdd
   *
   * @endif
   */
  typedef ConfigurationSetListener OnAddConfigurationAddCallback;

  /*!
   * @if jp
   * @class OnRemoveConfigurationSetCallback
   * @brief OnRemoveConfigurationSet R[obNۃNX
   *
   * @else
   * @class OnRemoveConfigurationSetCallback
   * @brief Callback functor abstract for OnRemoveConfigurationSet
   *
   * @endif
   */
  typedef ConfigurationSetNameListener OnRemoveConfigurationSetCallback;

  /*!
   * @if jp
   * @class OnActivateSetCallback
   * @brief OnActivateSet R[obNۃNX
   *
   * @else
   * @class OnActivateSetCallback
   * @brief Callback functor abstract for OnActivateSet
   *
   * @endif
   */
  typedef ConfigurationSetNameListener OnActivateSetCallback;

  // forward decl
  class ConfigAdmin;

  //============================================================
  // ConfigBase class
  //============================================================
  /*!
   * @if jp
   * @class ConfigBase
   * @brief ConfigBase ۃNX
   * 
   * eRtBM[Vێ邽߂̒ۃNXBۃR
   * tBM[VNX́Aȉ̏z֐̎񋟂Ȃ
   * ΂ȂȂB
   *
   * publicC^[tF[XƂĈȉ̂̂񋟂B
   * - update(): RtBM[Vp[^l̍XV
   *
   * @since 0.4.0
   *
   * @else
   * @class ConfigBase
   * @brief ConfigBase abstract class
   *
   * This is the abstract interface class to hold various configuration 
   * information.
   * Concrete configuration classes must implement the following pure virtual
   * functions.
   *
   * This class provides public interface as follows.
   * - update(): update configuration parameter value
   *
   * @since 0.4.0
   *
   * @endif
   */
  struct ConfigBase
  {
    /*!
     * @if jp
     *
     * @brief RXgN^
     * 
     * RXgN^
     *
     * @param name_ RtBM[V
     * @param def_val `̃ftHgl
     * 
     * @else
     *
     * @brief Constructer
     *
     * Constructer
     *
     * @param name_ Configuration name
     * @param def_val Default value in string format
     *
     * @endif
     */
    ConfigBase(const char* name_, const char* def_val)
      : name(name_), default_value(def_val),
        string_value(""), m_admin(NULL), m_callback(NULL)
    {}
    
    /*!
     * @if jp
     *
     * @brief zfXgN^
     * 
     * zfXgN^B
     *
     * @else
     *
     * @brief Virtual Destructor
     *
     * Virtual Destructor
     *
     * @endif
     */
    virtual ~ConfigBase(void){};

    // typedef of ConfigAdmin's member function
    typedef void (ConfigAdmin::*CallbackFunc)(const char*, const char*);

    /*!
     * @if jp
     *
     * @brief R[obÑZbg
     * 
     * ϐύXɃR[R[obN֐Zbg.
     *
     * @else
     *
     * @brief Setting callback
     *
     * This member function sets callback function which is called
     * when variable is changed.
     *
     * @endif
     */
    void setCallback(ConfigAdmin* cadmin, CallbackFunc cbf);

    /*!
     * @if jp
     *
     * @brief ϐύXm点IuU[o֐
     * 
     * ϐύXm点IuU[o֐.
     *
     * @else
     *
     * @brief Observer function to notify variable changed
     *
     * This function notifies variable has been changed.
     *
     * @endif
     */
    void notifyUpdate(const char* key, const char* val);
    
    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^lXVpz֐
     * 
     * RtBM[VݒlŃRtBM[Vp[^
     * XV邽߂̏z֐B
     *
     * @param val p[^l̕\
     *
     * @return ݒ茋
     * 
     * @else
     *
     * @brief Pure virtual function to update configuration parameter values
     * 
     * Pure virtual function to update configuration parameter 
     * by the configuration value.
     *
     * @param val The parameter values converted into character string format
     *
     * @return Result of the setup
     *
     * @endif
     */
    virtual bool update(const char* val) = 0;

    /*!
     * @if jp
     * @brief  RtBM[V
     * @else
     * @brief  Configuration name
     * @endif
     */
    const char* name;
    
    /*!
     * @if jp
     * @brief  `̃ftHgl
     * @else
     * @brief  Default value in string format
     * @endif
     */
    const char* default_value;

  protected:
    /*!
     * @if jp
     * @brief  `̌ݒl
     * @else
     * @brief  Current value in string format
     * @endif
     */
    std::string string_value;
    /*!
     * @if jp
     * @brief  ConfigAdminIuWFNgւ̃|C^
     * @else
     * @brief  A pointer to the ConfigAdmin object
     * @endif
     */
    ConfigAdmin* m_admin;
    /*!
     * @if jp
     * @brief  R[obÑo֐|C^
     * @else
     * @brief  A member function pointer to the callback function.
     * @endif
     */
    CallbackFunc m_callback;
  };
  
  //============================================================
  // Config template class
  //============================================================
  /*!
   * @if jp
   * @class Config
   * @brief Config NX
   * 
   * RtBM[Vp[^̏ێNXB
   * \<VarType\>ƂăRtBM[Ṽf[^^w肷B
   * \<TransFunc\>ƂĐݒ肳ꂽf[^^𕶎ɕϊϊ֐
   * w肷B
   *
   * @param VarType RtBM[Vp[^i[pϐ
   * @param TransFunc i[f[^^𕶎ɕϊϊ֐
   *
   * @since 0.4.0
   *
   * @else
   * @class Config
   * @brief Config class
   * 
   * Class to hold the configuration parameter information.
   * Specify the data type of the configuration as \<VarType\>
   * Specify transformation function to convert data type set as \<TransFunc\>
   * into string format.
   *
   * @param VarType Cariable to hold configuration parameter
   * @param TransFunc Transformation function to transform the stored data 
   * type into string format.
   *
   * @since 0.4.0
   *
   * @endif
   */
  template <typename VarType,
	    typename TransFunc = bool (*)(VarType&, const char*)>
  class Config
    : public ConfigBase
  {
  public:
    /*!
     * @if jp
     *
     * @brief RXgN^
     * 
     * RXgN^
     *
     * @param name RtBM[Vp[^
     * @param var RtBM[Vp[^i[pϐ
     * @param def_val `̃ftHgl
     * @param trans `ϊ֐
     * 
     * @else
     *
     * @brief Constructor
     * 
     * Constructor
     *
     * @param name Configuration parameter name
     * @param var Configuration parameter variable
     * @param def_val Default value in string format
     * @param trans Function to transform into string format
     *
     * @endif
     */
    Config(const char* name, VarType& var, const char* def_val,
	   TransFunc trans = coil::stringTo)
      : ConfigBase(name, def_val), m_var(var), m_trans(trans)
    {
    }
    
    /*!
     * @if jp
     *
     * @brief zfXgN^
     * 
     * zfXgN^B
     *
     * @else
     *
     * @brief Virtual Destructor
     * 
     * Virtual Destructor.
     *
     * @endif
     */
    virtual ~Config(void){}
    
    /*!
     * @if jp
     *
     * @brief oChp[^lXV
     * 
     * RtBM[VݒlŃRtBM[Vp[^XV
     *
     * @param val p[^l̕\
     *
     * @return XV(XV:trueCXVs:false)
     * 
     * @else
     *
     * @brief Update a bind parameter value
     * 
     * Update configuration paramater by the configuration value.
     *
     * @param val The parameter values converted into character string format
     *
     * @return Update result (Successful:true, Failed:false)
     *
     * @endif
     */
    virtual bool update(const char* val)
    {
      if (string_value == val) { return true; }
      string_value = val;
      // value changed
      if ((*m_trans)(m_var, val))
        {
          notifyUpdate(name, val);
          return true;
        }
      (*m_trans)(m_var, default_value);
      notifyUpdate(name, val);
      return false;
    }
    
  protected:
    /*!
     * @if jp
     * @brief  RtBM[Vp[^i[pϐ
     * @else
     * @brief  Configuration parameter variable
     * @endif
     */
    VarType& m_var;
    
    /*!
     * @if jp
     * @brief  RtBM[Vp[^^ϊ֐
     * @else
     * @brief  Transformation function to convert configuration parameter type 
     *         into string format.
     * @endif
     */
    TransFunc m_trans;
  };
  
  //============================================================
  // ConfigAdmin class
  //============================================================
  /*!
   * @if jp
   * @class ConfigAdmin
   * @brief ConfigAdmin NX
   * 
   * eRtBM[VǗNXB
   * pȉ̂悤ɒ`B
   *
   * - RtBM[V: R|[lg̐ݒB
   *
   * - (RtBM[V)p[^F key-value ȂݒB
   *   coil::Properties ϐƂĈAkeyAvalue ɕƂĕ
   *   Bkey RtBM[Vp[^Avalue R
   *   tBM[Vp[^lƌĂԁB
   *
   * - RtBM[VZbgF RtBM[Vp[^
   *   ̃XgŁAO (ID) ɂċʂBIDRtBM[V
   *   ZbgIDƌĂԁB
   *
   * - (RtBM[V)p[^ϐFRtBM[Vp
   *   [^RTC̃ANeBreBŎۂɗpۂɎQƂ
   *   Bp[^ƂɌŗĽ^B
   *
   * - ANeBu(RtBM[V)ZbgFݗLȃRtBM
   *   [VZbĝƂłAB݂BƂāAANeB
   *   uRtBM[VZbg̃p[^RtBM[V
   *   p[^ϐɔfB
   *
   * ̃NXł́ARtBM[V̂߂̈ȉ2̏
   * ĂB
   *
   * -# RtBM[VZbg̃Xg
   * -# p[^ϐ̃Xg
   *
   * {Iɂ́A(1) ̃RtBM[VZbg̃XĝA
   * (2) ̃p[^ϐ֔fÂ{NX̖ړIłBʏA
   * p[^ϐ̕ύX́ARtBM[VZbg̕ύXƃp
   * [^ϐւ̔f2iKōsB
   *
   * RtBM[VZbg̃Xg̑ɂ́Aȉ̊֐pB
   *
   * - getConfigurationSets()
   * - getConfigurationSet()
   * - setConfigurationSetValues()
   * - getActiveConfigurationSet()
   * - addConfigurationSet()
   * - removeConfigurationSet()
   * - activateConfigurationSet()
   *
   * ̊֐ɂARtBM[VZbg̕ύXAǉA폜A
   * 擾AANeBusB̑ɂύXꂽRtBM
   * [VZbgARTC̃ANeBreBgpp[^ϐ
   * ɔfɂ́Aȉ update() ֐pB
   *
   * - update(void)
   * - update(const char* config_set)
   * - update(const char* config_set, const char* config_param)
   *
   * RtBM[VtbN邽߂ɃR[obNt@N^
   * ^邱ƂłBtbNł鑀͈ȉ̒ʂB
   *
   * - ON_UPDATE                   : update() R[
   * - ON_UPDATE_PARAM             : update(param) R[
   * - ON_SET_CONFIGURATIONSET     : setConfigurationSet() R[
   * - ON_ADD_CONFIGURATIONSET     : addConfigurationSet() R[
   * - ON_REMOVE_CONFIGURATIONSET  : removeConfigurationSet() R[
   * - ON_ACTIVATE_CONFIGURATIONSET: activateConfigurationSet() R[
   *
   * @since 0.4.0
   *
   * @else
   * @class ConfigAdmin
   * @brief ConfigAdmin class
   * 
   * Class to manage various configuration information.
   * Now terms for this class are defined as follows.
   *
   * - Configurations: The configuration information for the RTCs.
   *
   * - (Configuration) parameters: Configuration information that
   *   consists of a key-value pair. The "key" and the "value" are
   *   both stored as character string values in a coil::Properties
   *   variable in this class. The "key" is called the "configuration
   *   parameter name", and the "value" is called the "configuration
   *   parameter value".
   *
   * - Configuration-sets: This is a list of configuration parameters,
   *   and it is distinguished by name (ID). The ID is called
   *   configuration-set ID.
   *
   * - (Configuration) parameter variables: The variables to be
   *   referred when configuration parameters are actually used within
   *   the activity of an RTC. Each variable has each type.
   *
   * - Active (configuration) set: This is the only configuration-set
   *   that is currently active. The parameter values of the active
   *    configuration-set are substituted into configuration variables
   *   in principle.
   *
   * The following two configuration informations are stored in this class.
   *
   * -# A list of configuration-set
   * -# A list of configuration parameter variables
   *
   * Basically, the purpose of this class is to set one of the
   * configuration-set in the list of (1) into parameter variables of
   * (2). Usually, configuration parameter variables manipulation is
   * performed with two-phases of configuration-set setting and
   * parameter variables setting.
   *
   * The configuration-set manipulations are performed by the
   * following functions.
   *
   * - getConfigurationSets()
   * - getConfigurationSet()
   * - setConfigurationSetValues()
   * - getActiveConfigurationSet()
   * - addConfigurationSet()
   * - removeConfigurationSet()
   * - activateConfigurationSet()
   *
   * Modification, addition, deletion, acquisition and activation of
   * configuration-set are performed by these functions. In order to
   * reflect configuration-set, which is manipulated by these
   * functions, on parameter variables that are used from RTC
   * activities, the following update() functions are used .
   *
   * - update(void)
   * - update(const char* config_set)
   * - update(const char* config_set, const char* config_param)
   *
   * Callback functors can be given to hook configuration
   * operation. Operations to be hooked are as follows.
   *
   * - ON_UPDATE                   : when update() is called
   * - ON_UPDATE_PARAM             : when update(param) is called
   * - ON_SET_CONFIGURATIONSET     : when setConfigurationSet() is called
   * - ON_ADD_CONFIGURATIONSET     : when addConfigurationSet() is called
   * - ON_REMOVE_CONFIGURATIONSET  : when removeConfigurationSet() is called
   * - ON_ACTIVATE_CONFIGURATIONSET: when activateConfigurationSet() is called
   *
   * @since 0.4.0
   *
   * @endif
   */
  class ConfigAdmin
  {
  public:
    /*!
     * @if jp
     *
     * @brief RXgN^
     * 
     * RXgN^
     *
     * @param prop ݒΏۃvpeB
     * 
     * @else
     *
     * @brief Constructor
     * 
     * Constructor
     *
     * @param prop The target property name for setup
     *
     * @endif
     */
    ConfigAdmin(coil::Properties& prop);
    
    /*!
     * @if jp
     *
     * @brief zfXgN^
     * 
     * zfXgN^B
     *
     * @else
     *
     * @brief Virtual Destructor
     * 
     * Virtual Destructor
     *
     * @endif
     */
    ~ConfigAdmin(void);
    
    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̐ݒ
     * 
     * RtBM[Vp[^ƕϐoCh
     * w肵̂̃RtBM[Vp[^ɑ݂ꍇ
     * falseԂB
     * \<VarType\>ƂăRtBM[Vp[^̃f[^^w肷B
     * ȂÃo֐Ă΂ꂽA݃ANeBuȃRtBM[
     * VZbg̒lϐɃZbgB
     *
     * @param param_name RtBM[Vp[^
     * @param var RtBM[Vp[^i[pϐ
     * @param def_val RtBM[Vp[^ftHgl
     * @param trans RtBM[Vp[^ϊp֐
     *
     * @return ݒ茋(ݒ萬:trueCݒ莸s:false)
     * 
     * @else
     *
     * @brief Setup for configuration parameters
     * 
     * Bind configuration parameter to its variable.
     * Return false, if configuration parameter of specified name has already 
     * existed.
     * Specify the data type of the configuration as \<VarType\>.
     * After calling this function, configuration variable would be
     * updated by the active configuration parameter.
     *
     * @param param_name Configuration parameter name
     * @param var Configuration parameter variable
     * @param def_val Default value of configuration parameter
     * @param trans Function to transform configuration parameter type into 
     *        string format
     *
     * @return Setup result (Successful:true, Failed:false)
     *
     * @endif
     */
    template <typename VarType>
    bool bindParameter(const char* param_name, VarType& var,
		       const char* def_val,
		       bool (*trans)(VarType&, const char*) = coil::stringTo)
    {
      if (param_name == 0) { return false; }
      if (def_val == 0) { return false; }
      if (isExist(param_name)) { return false; }
      if (!trans(var, def_val)) { return false; }
      Config<VarType>* c = new Config<VarType>(param_name, var, def_val, trans);
      m_params.push_back(c);
      c->setCallback(this, &RTC::ConfigAdmin::onUpdateParam);
      update(getActiveId(), param_name);
      return true;
    }

    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̉
     * 
     * RtBM[Vp[^ƕϐ̃oChB
     * w肵̂̃RtBM[Vp[^݂Ȃꍇ
     * falseԂB
     *
     * @param param_name RtBM[Vp[^
     * @return ݒ茋(ݒ萬:trueCݒ莸s:false)
     * 
     * @else
     *
     * @brief Unbinding configuration parameters
     * 
     * Unbind configuration parameter from its variable. It returns
     * false, if configuration parameter of specified name has already
     * existed.
     *
     * @param param_name Configuration parameter name
     * @return Setup result (Successful:true, Failed:false)
     *
     * @endif
     */
    bool unbindParameter(const char* param_name);
        
    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̍XV
     *        (ANeBuRtBM[VZbg)
     * 
     * RtBM[VZbgXVĂꍇɁA݃ANeB
     * uɂȂĂRtBM[Vɐݒ肵lŁARtBM
     * [Vp[^̒lXVB̏ł̍XV́AANeB
     * uƂȂĂRtBM[VZbg݂ĂꍇAO
     * ̍XVRtBM[VZbg̓eXVĂ
     * ̂ݎsB
     *
     * @else
     *
     * @brief Update the values of configuration parameters
     *        (Active configuration set)
     * 
     * When configuration set is updated, update the configuration
     * parameter value to the value that is set to the current active
     * configuration.  This update will be executed, only when an
     * active configuration set exists and the content of the
     * configuration set has been updated from the last update.
     *
     * @endif
     */
    void update(void);

    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̍XV(IDw)
     * 
     * RtBM[Vϐ̒lAw肵IDRtBM[
     * VZbg̒lōXVBɂAANeBuȃRtBM
     * [VZbg͕ύXȂBāAANeBuRtBM
     * [VZbgƃp[^ϐ̊Ԃɖ\
     * ̂ŒӂKvłB
     *
     * w肵ID̃RtBM[VZbg݂Ȃꍇ́A
     * ɏIB
     *
     * @param config_set ݒΏۂ̃RtBM[VZbgID
     * 
     * @else
     *
     * @brief Update configuration parameter (By ID)
     * 
     * This operation updates configuration variables by the
     * configuration-set with specified ID. This operation does not
     * change current active configuration-set. Since this operation
     * causes inconsistency between current active configuration set
     * and actual values of configuration variables, user should
     * carefully use it.
     *
     * This operation ends without doing anything, if the
     * configuration-set does not exist.
     *
     * @param config_set The target configuration set's ID to setup
     *
     * @endif
     */
    void update(const char* config_set);
    
    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̍XV(̎w)
     * 
     * ̃RtBM[Vϐ̒lAw肵IDRtB
     * M[VZbg̒lōXVBɂAANeBuȃR
     * tBM[VZbg͕ύXȂBāAANeBuR
     * tBM[VZbgƃp[^ϐ̊Ԃɖ
     * \̂ŒӂKvłB
     *
     * w肵ID̃RtBM[VZbgAw肵̂̃p[
     * ^݂Ȃꍇ́AɏIB
     *
     * @param config_set RtBM[VID
     * @param config_param RtBM[Vp[^
     * 
     * @else
     *
     * @brief Update the values of configuration parameters (By name)
     * 
     * This operation updates a configuration variable by the
     * specified configuration parameter in the
     * configuration-set. This operation does not change current
     * active configuration-set. Since this operation causes
     * inconsistency between current active configuration set and
     * actual values of configuration variables, user should carefully
     * use it.
     *
     * This operation ends without doing anything, if the
     * configuration-set or the configuration parameter do not exist.
     *
     * @param config_set configuration-set ID.
     * @param config_param configuration parameter name.
     *
     * @endif
     */
    void update(const char* config_set, const char* config_param);
    
    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̑݊mF
     * 
     * w肵̂RtBM[Vp[^ϐ݂
     * mFBő݊mFsp[^ϐƂ́A
     * bindParameter() ɂēo^Aϐp[^łB
     *
     * @param name RtBM[Vp[^́B
     *
     * @return ݊mF(p[^:trueCp[^Ȃ:false)
     *
     * @else
     *
     * @brief Check the existence of configuration parameters
     * 
     * Check the existence of configuration parameters of specified name.
     *
     * @param name Configuration parameter name
     *
     * @return Result of existance confirmation 
     *         (Parameters exist:true, else:false)
     *
     * @endif
     */
    bool isExist(const char* name);
    
    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̕ύXmF
     * 
     * RtBM[Vp[^ύXꂽmFB
     *
     * @return ύXmF(ύX:trueAύXȂ:false)
     *
     * @else
     *
     * @brief Confirm to change configuration parameters
     * 
     * Confirm that configuration parameters have changed.
     *
     * @return Result of change confirmation
     *         (There is a change:trueANo change:false)
     *
     * @endif
     */
    bool isChanged(void) {return m_changed;}

    /*!
     * @if jp
     *
     * @brief ύXꂽp[^̃Xg
     * 
     * RtBM[Vp[^̂ύXꂽ̂̃XgԂB
     *
     * @return ύXꂽp[^Xg
     *
     * @else
     *
     * @brief Changed parameters list
     * 
     * This operation returns parameter list which are changed.
     *
     * @return Changed parameters list
     *
     * @endif
     */
    coil::vstring& changedParameters() { return m_changedParam; }
    
    /*!
     * @if jp
     *
     * @brief ANeBuERtBM[VZbgID̎擾
     * 
     * ݃ANeBuȃRtBM[VZbgID擾B
     *
     * @return ANeBuERtBM[VZbgID
     *
     * @else
     *
     * @brief Get ID of active configuration set
     * 
     * Get ID of the current active configuration set.
     *
     * @return The active configuration set ID
     *
     * @endif
     */
    const char* getActiveId(void) {return m_activeId.c_str();}
    
    /*!
     * @if jp
     *
     * @brief RtBM[VZbg̑݊mF
     * 
     * w肵RtBM[VZbg݂邩mFB
     *
     * @param config_id mFΏۃRtBM[VZbgID
     *
     * @return ݊mF(w肵ConfigSet:trueAȂ:false)
     *
     * @else
     *
     * @brief Check the existence of configuration set
     * 
     * Check the existence of specified configuration set.
     *
     * @param config_id ID of target configuration set for confirmation
     *
     * @return Result of existence confirmation 
     *         (Specified ConfigSet exists:true, else:false)
     *
     * @endif
     */
    bool haveConfig(const char* config_id)
    {
      return (m_configsets.hasKey(config_id) == NULL) ? false : true;
    }
    
    /*!
     * @if jp
     *
     * @brief RtBM[VZbg̃ANeBumF
     * 
     * RtBM[VZbgANeBuĂ邩mFB
     *
     * @return ԊmF(ANeBu:trueAANeBu:false)
     *
     * @else
     *
     * @brief Confirm to activate configuration set
     * 
     * Confirm that configuration set has been activated.
     *
     * @return Result of state confirmation
     *         (Active state:true, Inactive state:false)
     *
     * @endif
     */
    bool isActive(void)
    {
      return m_active;
    }
    //    const std::vector<Properties*>* getConfigurationParameterValues();
    //    const Properties* getConfigurationParameterValue(const char* name);
    //    bool setConfigurationParameter(const char* name, const char* value);
    
    /*!
     * @if jp
     *
     * @brief SRtBM[VZbg̎擾
     * 
     * ݒ肳ĂSRtBM[VZbg擾B
     *
     * @return SRtBM[VZbg
     *
     * @else
     *
     * @brief Get all configuration sets
     * 
     * Get all specified configuration sets
     *
     * @return All configuration sets
     *
     * @endif
     */
    const std::vector<coil::Properties*>& getConfigurationSets(void);
    
    /*!
     * @if jp
     *
     * @brief w肵ID̃RtBM[VZbg̎擾
     * 
     * IDŎw肵RtBM[VZbg擾B
     * w肵RtBM[VZbg݂Ȃꍇ́A
     * ̃RtBM[VZbgԂB
     *
     * @param config_id 擾ΏۃRtBM[VZbgID
     *
     * @return RtBM[VZbg
     *
     * @else
     *
     * @brief Get a configuration set by specified ID
     * 
     * Get a configuration set that was specified by ID
     * Return empty configuration set, if a configuration set of
     * specified ID doesn't exist.
     *
     * @param config_id ID of the target configuration set for getting
     *
     * @return The configuration set
     *
     * @endif
     */
    const coil::Properties& getConfigurationSet(const char* config_id);
    
    /*!
     * @if jp
     *
     * @brief w肵vpeB̃RtBM[VZbgւ̒ǉ
     * 
     * w肵vpeBIDŎw肵RtBM[VZbg֒ǉB
     * w肵IDƈvRtBM[VZbg݂Ȃꍇ́A
     * false ԂB
     *
     * @param config_id ǉΏۃRtBM[VZbgID
     * @param configuration_set ǉvpeB
     *
     * @return ǉs(ǉ:trueAǉs:false)
     *
     * @else
     *
     * @brief Add to configuration set from specified property
     * 
     * Add specified property to configuration set that was specified by ID.
     * Return false if configuration set, that matches specified ID, 
     * doesn't exist.
     *
     * @param config_id ID of the target configuration set for add
     * @param configuration_set Property to add
     *
     * @return Add result (Successful:true, Failed:false)
     *
     * @endif
     */
    bool setConfigurationSetValues(const coil::Properties& configuration_set);
    
    /*!
     * @if jp
     *
     * @brief ANeBuERtBM[VZbg擾
     * 
     * ݃ANeBuƂȂĂRtBM[VZbg擾B
     * ANeBuƂȂĂRtBM[VZbg݂Ȃꍇ́A
     * ̃RtBM[VZbg ԂB
     *
     * @return ANeBuERtBM[VZbg
     *
     * @else
     *
     * @brief Get the active configuration set
     * 
     * Get the current active configuration set.
     * Return empty configuration set, if an active configuration set 
     * doesn't exist.
     *
     * @return The active configuration set
     *
     * @endif
     */
    const coil::Properties& getActiveConfigurationSet(void);
    
    /*!
     * @if jp
     *
     * @brief RtBM[VZbgɐݒlǉ
     * 
     * RtBM[VZbgɐݒlǉB
     *
     * @param configuration_set ǉvpeB
     *
     * @return ǉ(ǉ:trueAǉs:false)
     *
     * @else
     *
     * @brief Add the configuration value to configuration set
     * 
     * Add the configuration value to configuration set
     *
     * @param configuration_set Property to add
     *
     * @return Add Result (Successful:true, Failed:false)
     *
     * @endif
     */
    bool addConfigurationSet(const coil::Properties& configuration_set);
    
    /*!
     * @if jp
     *
     * @brief RtBM[VZbg̍폜
     * 
     * w肵ID̃RtBM[VZbg폜B
     *
     * w肵ID̃RtBM[VZbg݂Ȃꍇ́A
     * falseԂB폜\ȃRtBM[VZbǵA
     * addConfigruationSet() ɂĒǉRtBM[VZb
     * ĝ݂łAftHgRtBM[VZbgAR|[l
     * gNɃt@Cǂݍ܂RtBM[VZbg
     * ͍폜邱ƂłȂB
     *
     * ܂Aw肵RtBM[VZbg݃ANeBuł
     * ꍇɂ́AȂRtBM[VZbgł폜łȂB
     *
     * ̊֐ɂۂɃRtBM[VZbg폜ꂽꍇA
     * setOnRemoveConfigurationSet() ŃZbgꂽR[obN֐
     * яoB
     *
     * @param config_id 폜ΏۃRtBM[VZbgID
     *
     * @return 폜(폜:trueA폜s:false)
     *
     * @else
     *
     * @brief Remove the configuration set
     * 
     * Remove the configuration set of specified ID Return empty
     * configuration set, if a configuration set of specified ID
     * doesn't exist.
     *
     * The configuration-sets that can be removed by this function are
     * only configuration-sets newly added by the
     * addConfigurationSet() function. The configuration that can be
     * removed by this function is only newly added configuration-set
     * by addConfigurationSet() function.  The "default"
     * configuration-set and configurationi-sets that is loaded from
     * configuration file cannot be removed.
     *
     * If the specified configuration is active currently, any
     * configurations are not deleted.
     *
     * Callback functions that are set by
     * addOnRemovedConfigurationSet() will be called if a
     * configuration-set is deleted actually by this function.
     *
     * @param config_id ID of the target configuration set for remove
     *
     * @return Remove result (Successful:true, Failed:false)
     *
     * @endif
     */
    bool removeConfigurationSet(const char* config_id);
    
    /*!
     * @if jp
     *
     * @brief RtBM[VZbg̃ANeBu
     * 
     * w肵ID̃RtBM[VZbgANeBuB
     * w肵ID̃RtBM[VZbg݂Ȃꍇ́A
     * falseԂB
     *
     * @param config_id 폜ΏۃRtBM[VZbgID
     *
     * @return ANeBu(:trueAs:false)
     *
     * @else
     *
     * @brief Activate the configuration set
     * 
     * Activate the configuration set of specified ID
     * Return empty configuration set, if a configuration set of
     * specified ID doesn't exist.
     *
     * @param config_id ID of the target configuration set for remove
     *
     * @return Activate result (Remove success:trueARemove failure:false)
     *
     * @endif
     */
    bool activateConfigurationSet(const char* config_id);

    //------------------------------------------------------------
    // obsolete functions
    //
    void setOnUpdate(OnUpdateCallback* cb);
    
    void setOnUpdateParam(OnUpdateParamCallback* cb);
    
    void setOnSetConfigurationSet(OnSetConfigurationSetCallback* cb);

    void setOnAddConfigurationSet(OnAddConfigurationAddCallback* cb);

    void setOnRemoveConfigurationSet(OnRemoveConfigurationSetCallback* cb);

    void setOnActivateSet(OnActivateSetCallback* cb);
    //
    // end of obsolete functions
    //------------------------------------------------------------

    /*!
     * @if jp
     *
     * @brief ConfigurationParamListener ǉ
     *
     * update(const char* config_set, const char* config_param) Ă΂ꂽۂ
     * R[郊Xi ConfigurationParamListener ǉB
     * type ɂ݂͌̂Ƃ ON_UPDATE_CONFIG_PARAM ݂̂B
     *
     * @param type ConfigurationParamListenerType^̒lB
     *             ON_UPDATE_CONFIG_PARAM B
     *
     * @param listener ConfigurationParamListener ^̃XiIuWFNgB
     * @param autoclean XiIuWFNgō폜邩ǂ̃tO
     * 
     * @else
     *
     * @brief Adding ConfigurationParamListener 
     * 
     * This function adds a listener object which is called when
     * update(const char* config_set, const char* config_param) is
     * called. In the type argument, currently only
     * ON_UPDATE_CONFIG_PARAM is allowed.
     *
     * @param type ConfigurationParamListenerType value
     *             ON_UPDATE_CONFIG_PARAM is only allowed.
     *
     * @param listener ConfigurationParamListener listener object.
     * @param autoclean a flag whether if the listener object autocleaned.
     *
     * @endif
     */
    void addConfigurationParamListener(ConfigurationParamListenerType type,
                                       ConfigurationParamListener* listener,
                                       bool autoclean = true);

    /*!
     * @if jp
     *
     * @brief ConfigurationParamListener 폜
     *
     * addConfigurationParamListener ŒǉꂽXiIuWFNg폜B
     *
     * @param type ConfigurationParamListenerType^̒lB
     *             ON_UPDATE_CONFIG_PARAM B
     * @param listener ^XiIuWFNgւ̃|C^
     * 
     * @else
     *
     * @brief Removing ConfigurationParamListener 
     * 
     * This function removes a listener object which is added by
     * addConfigurationParamListener() function.
     *
     * @param type ConfigurationParamListenerType value
     *             ON_UPDATE_CONFIG_PARAM is only allowed.
     * @param listener a pointer to ConfigurationParamListener listener object.
     *
     * @endif
     */
    void removeConfigurationParamListener(ConfigurationParamListenerType type,
                                          ConfigurationParamListener* listener);
    
    /*!
     * @if jp
     *
     * @brief ConfigurationSetListener ǉ
     *
     * ConfigurationSet XVꂽƂȂǂɌĂ΂郊Xi
     * ConfigurationSetListener ǉBݒ\ȃCxg͈ȉ
     * 2ނB
     *
     * - ON_SET_CONFIG_SET: setConfigurationSetValues() 
     *                      ConfigurationSet ɒlݒ肳ꂽꍇB
     * - ON_ADD_CONFIG_SET: addConfigurationSet() ŐV
     *                      ConfigurationSet ǉꂽꍇB
     *
     * @param type ConfigurationSetListenerType^̒lB
     * @param listener ConfigurationSetListener ^̃XiIuWFNgB
     * @param autoclean XiIuWFNgō폜邩ǂ̃tO
     * 
     * @else
     *
     * @brief Adding ConfigurationSetListener 
     * 
     * This function add a listener object which is called when
     * ConfigurationSet is updated. Available events are the followings.
     *
     * @param type ConfigurationSetListenerType value
     * @param listener ConfigurationSetListener listener object.
     * @param autoclean a flag whether if the listener object autocleaned.
     *
     * @endif
     */
    void addConfigurationSetListener(ConfigurationSetListenerType type,
                                     ConfigurationSetListener* listener,
                                     bool autoclean = true);

    /*!
     * @if jp
     *
     * @brief ConfigurationSetListener 폜
     *
     * addConfigurationSetListener ŒǉꂽXiIuWFNg폜B
     *
     * @param type ConfigurationSetListenerType^̒lB
     * @param listener ^XiIuWFNgւ̃|C^
     * 
     * @else
     *
     * @brief Removing ConfigurationSetListener 
     * 
     * This function removes a listener object which is added by
     * addConfigurationSetListener() function.
     *
     * @param type ConfigurationSetListenerType value
     * @param listener a pointer to ConfigurationSetListener listener object.
     *
     * @endif
     */
    void removeConfigurationSetListener(ConfigurationSetListenerType type,
                                        ConfigurationSetListener* listener);
    
    /*!
     * @if jp
     *
     * @brief ConfigurationSetNameListener ǉ
     *
     * ConfigurationSetName XVꂽƂȂǂɌĂ΂郊Xi
     * ConfigurationSetNameListener ǉBݒ\ȃCxg͈ȉ
     * 3ނB
     *
     * - ON_UPDATE_CONFIG_SET:  ConfigurationSet Abvf[gꂽ
     * - ON_REMOVE_CONFIG_SET:  ConfigurationSet 폜ꂽ
     * - ON_ACTIVATE_CONFIG_SET:  ConfigurationSet ANeBuꂽ
     *
     * @param type ConfigurationSetNameListenerType^̒lB
     * @param listener ConfigurationSetNameListener ^̃XiIuWFNgB
     * @param autoclean XiIuWFNgō폜邩ǂ̃tO
     * 
     * @else
     *
     * @brief Adding ConfigurationSetNameListener 
     * 
     * This function add a listener object which is called when
     * ConfigurationSetName is updated. Available events are the followings.
     *
     * - ON_UPDATE_CONFIG_SET: A ConfigurationSet has been updated.
     * - ON_REMOVE_CONFIG_SET: A ConfigurationSet has been deleted.
     * - ON_ACTIVATE_CONFIG_SET: A ConfigurationSet has been activated.
     *
     * @param type ConfigurationSetNameListenerType value
     * @param listener ConfigurationSetNameListener listener object.
     * @param autoclean a flag whether if the listener object autocleaned.
     *
     * @endif
     */
    void 
    addConfigurationSetNameListener(ConfigurationSetNameListenerType type,
                                    ConfigurationSetNameListener* listener,
                                    bool autoclean = true);

    /*!
     * @if jp
     *
     * @brief ConfigurationSetNameListener 폜
     *
     * addConfigurationSetNameListener ŒǉꂽXiIuWFNg
     * 폜B
     *
     * @param type ConfigurationSetNameListenerType^̒lB
     *             ON_UPDATE_CONFIG_PARAM B
     * @param listener ^XiIuWFNgւ̃|C^
     * 
     * @else
     *
     * @brief Removing ConfigurationSetNameListener 
     * 
     * This function removes a listener object which is added by
     * addConfigurationSetNameListener() function.
     *
     * @param type ConfigurationSetNameListenerType value
     *             ON_UPDATE_CONFIG_PARAM is only allowed.
     * @param listener a pointer to ConfigurationSetNameListener
     *             listener object.
     *
     * @endif
     */
    void
    removeConfigurationSetNameListener(ConfigurationSetNameListenerType type,
                                       ConfigurationSetNameListener* listener);
    
  protected:
    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̍XV(IDw)ɃR[
     *
     * ݒ肳ĂR[obNIuWFNgĂяoB
     *
     * @param config_set ݒΏۂ̃RtBM[VZbgID
     *
     * @else
     *
     * @brief When the configuration parameter is updated, it is called. 
     *
     * Call the set callback object.
     * 
     * @param config_set The target configuration set's ID to setup
     *
     * @endif
     */
    void onUpdate(const char* config_set);

    /*!
     * @if jp
     *
     * @brief RtBM[Vp[^̍XV(̎w)ɃR[
     *
     * ݒ肳ĂR[obNIuWFNgĂяoB
     *
     * @param config_set RtBM[VID
     * @param config_param RtBM[Vp[^
     *
     * @else
     *
     * @brief When the configuration parameter is updated, it is called. 
     *
     * Call the set callback object.
     * 
     * @param config_set configuration-set ID.
     * @param config_param configuration parameter name.
     *
     * @endif
     */
    void onUpdateParam(const char* config_set, const char* config_param);

    /*!
     * @if jp
     *
     * @brief RtBM[VZbgւ̒ǉɃR[
     *
     * ݒ肳ĂR[obNIuWFNgĂяoB
     *
     * @param configuration_set vpeB
     *
     * @else
     *
     * @brief Called when the property is added to the configuration set
     *
     * Call the set callback object.
     * 
     * @param configuration_set property
     *
     * @endif
     */
    void onSetConfigurationSet(const coil::Properties& config_set);

    /*!
     * @if jp
     *
     * @brief ݒlǉꂽƂɃR[B
     *
     * ݒ肳ĂR[obNIuWFNgĂяoB
     *
     * @param configuration_set vpeB
     *
     * @else
     *
     * @brief Called when a set value is added to the configuration set
     *
     * Call the set callback object.
     * 
     * @param configuration_set property
     *
     * @endif
     */
    void onAddConfigurationSet(const coil::Properties& config_set);

    /*!
     * @if jp
     *
     * @brief Zbg폜ĂƂɃR[B
     *
     * ݒ肳ĂR[obNIuWFNgĂяoB
     *
     * @param config_id vpeB
     *
     * @else
     *
     * @brief Called when the configuration set has been deleted
     *
     * Call the set callback object.
     * 
     * @param config_id property
     *
     * @endif
     */
    void onRemoveConfigurationSet(const char* config_id);

    /*!
     * @if jp
     *
     * @brief ZbgANeBuꂽƂɃR[B
     *
     * ݒ肳ĂR[obNIuWFNgĂяoB
     *
     * @param config_id vpeB
     *
     * @else
     *
     * @brief Called when the configuration set is made active
     *
     * Call the set callback object.
     * 
     * @param config_id property
     *
     * @endif
     */
    void onActivateSet(const char* config_id);
    
  private:
    ConfigAdmin(const ConfigAdmin& ca);
    ConfigAdmin& operator=(const ConfigAdmin& ca);
    
    struct find_conf
    {
      std::string m_name;
      find_conf(const char* name) : m_name(name) {};
      bool operator()(ConfigBase* conf)
      {
        if (conf == 0) { return false; }
	return (m_name == conf->name);
      }
    };
    
    coil::Properties& m_configsets;
    coil::Properties  m_emptyconf;
    std::vector<ConfigBase*> m_params;
    std::string m_activeId;
    bool m_active;
    bool m_changed;
    coil::vstring m_changedParam;
    coil::vstring m_newConfig;
    ConfigurationListeners m_listeners;

  };
}; // namespace RTC
#endif // RTC_CONFIGADMIN_H
