// -*- C++ -*-
/*!
 * @file ExecutionContextWorker.h
 * @brief ExecutionContext's state machine worker class
 * @date $Date$
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2011
 *     Noriaki Ando
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id$
 *
 */

#ifndef RTC_EXECUTIONCONTEXTWORKER_H
#define RTC_EXECUTIONCONTEXTWORKER_H

#include <coil/Mutex.h>
#include <coil/Condition.h>
#include <vector>

#include <rtm/idl/RTCSkel.h>
#include <rtm/SystemLogger.h>
//#include <rtm/StateMachine.h>
//#include <rtm/RTObjectStateMachine.h>

#define NUM_OF_LIFECYCLESTATE 4

#ifdef WIN32
#pragma warning( disable : 4290 )
#endif

namespace RTC
{
  class RTObject_impl;
};
namespace RTC_impl
{
  class RTObjectStateMachine;
  typedef RTC::LightweightRTObject_ptr LightweightRTObject_ptr;
  typedef RTC::LightweightRTObject_var LightweightRTObject_var;
  /*!
   * @if jp
   * @class ExecutionContextWorker
   * @brief ExecutionContextWorker NX
   *
   * RTC̏ԃ}VEvLVNX RTObjectStateMachine ێEǗ
   * AR[obNۂɃR[邽߂̃[J[NXC^[tF[
   * XNXBEC̎҂́AExecutionContextBase pAoϐ
   * ƂĕێĂ邱 ExecutionContextWorker NX̃C^[
   * tF[XoRRTObject̃R[obNR[邱ƂɂȂB
   *
   * - EC̃IuWFNgt@X̃Zb^EQb^[
   *   - void setECRef(RTC::ExecutionContextService_ptr ref);
   *   - RTC::ExecutionContextService_ptr getECRef();
   *   EC̋ۃNXɂ́AANeBuEC̃IuWFNgt@
   *   XsetECRef() ŃZbgB
   *
   * - EC̎sE~֌W֐
   *   - CORBA::Boolean isRunning(void);
   *   - RTC::ReturnCode_t start(void);
   *   - RTC::ReturnCode_t stop(void);
   *
   * - RTC̃ANeBuEANeBu֐
   *   - RTC::ReturnCode_t activateComponent(RTC::LightweightRTObject_ptr comp,
   *                                         RTObjectStateMachine*& rtobj);
   *   - RTC::ReturnCode_t waitActivateComplete(RTObjectStateMachine*& rtobj,
   *                                            coil::TimeValue timeout = 1.0,
   *                                            long int cycle = 1000);
   *
   *   - RTC::ReturnCode_t deactivateComponent(RTC::LightweightRTObject_ptr comp,  *                                           RTObjectStateMachine*& rtobj);
   *   - RTC::ReturnCode_t waitDeactivateComplete(RTObjectStateMachine*& rtobj,
   *                                              coil::TimeValue timeout = 1.0,
   *                                              long int cycle = 1000);
   *   -  RTC::ReturnCode_t resetComponent(RTC::LightweightRTObject_ptr com,
   *                                       RTObjectStateMachine*& rtobj);
   *   -  RTC::ReturnCode_t waitResetComplete(RTObjectStateMachine*& rtobj,
   *                                          coil::TimeValue timeout = 1.0,
   *                                          long int cycle = 1000);
   *   - RTC::LifeCycleState getComponentState(RTC::LightweightRTObject_ptr comp);
   *   - const char* getStateString(RTC::LifeCycleState state)
   *
   * - RTCǉE폜n֐
   *   - RTC::ReturnCode_t addComponent(RTC::LightweightRTObject_ptr comp);
   *   - RTC::ReturnCode_t bindComponent(RTC::RTObject_impl* rtc);
   *   - RTC::ReturnCode_t removeComponent(RTC::LightweightRTObject_ptr comp);
   *   - RTObjectStateMachine* findComponent(RTC::LightweightRTObject_ptr comp);
   *
   * - Ԗ₢킹n֐
   *   - bool isAllCurrentState(RTC::LifeCycleState state);
   *   - bool isAllNextState(RTC::LifeCycleState state);
   *   - bool isOneOfCurrentState(RTC::LifeCycleState state);
   *   - bool isOneOfNextState(RTC::LifeCycleState state);
   *
   * - Worker̋쓮уR[obN
   *   - void invokeWorker();
   *   - void invokeWorkerPreDo();
   *   - void invokeWorkerDo();
   *   - void invokeWorkerPostDo();
   *
   * @since 1.2
   *
   * @else
   * @class ExecutionContextWorker
   * @brief ExecutionContextWorker class
   *
   * Periodic Sampled Data Processing (for the execution cycles)
   * ExecutionContext class
   *
   * @since 0.4.0
   *
   * @endif
   */
  class ExecutionContextWorker
  {
    typedef coil::Guard<coil::Mutex> Guard;

  public:
    /*!
     * @if jp
     * @brief ftHgRXgN^
     *
     * ftHgRXgN^
     * vt@CɈȉ̍ڂݒ肷B
     *  - kind : PERIODIC
     *  - rate : 0.0
     *
     * @else
     * @brief Default Constructor
     *
     * Default Constructor
     * Set the following items to profile.
     *  - kind : PERIODIC
     *  - rate : 0.0
     *
     * @endif
     */
    ExecutionContextWorker();


    /*!
     * @if jp
     * @brief fXgN^
     *
     * fXgN^
     *
     * @else
     * @brief Destructor
     *
     * Destructor
     *
     * @endif
     */
    virtual ~ExecutionContextWorker(void);

    //============================================================
    // Object reference to EC
    //============================================================
    void setECRef(RTC::ExecutionContextService_ptr ref);
    RTC::ExecutionContextService_ptr getECRef();

    //============================================================
    // ExecutionContext
    //============================================================
    /*!
     * @if jp
     * @brief ExecutionContext sԊmF֐
     *
     * ̑ ExecutionContext  Runnning Ԃ̏ꍇ true ԂB
     * Executioncontext  Running ̊ԁAY Executioncontext ɎQ
     * ĂSẴANeBuRTR|[lgAExecutionContext ̎
     * sނɉĎsB
     *
     * @return ԊmF֐(쒆:trueA~:false)
     *
     * @else
     *
     * @brief Check for ExecutionContext running state
     *
     * This operation shall return true if the context is in the
     * Running state.  While the context is Running, all Active RTCs
     * participating in the context shall be executed according to the
     * contextfs execution kind.
     *
     * @return Check state function (Running:trueAStopping:false)
     *
     * @endif
     */
    CORBA::Boolean isRunning(void);

    /*!
     * @if jp
     * @brief ExecutionContext ̎sJn
     *
     * ExecutionContext ̎sԂ Runnning Ƃ邽߂̃NGXg
     * sBExecutionContext ̏ԂJڂ
     * ComponentAction::on_startup ĂяoBQĂRTR|[
     * lgA܂ ExecutionContext Jn邱Ƃ͂ł
     * ȂBExecutionContext ͕Jn/~JԂƂłB
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Start the ExecutionContext
     *
     * Request that the context enter the Running state.  Once the
     * state transition occurs, the ComponentAction::on_startup
     * operation will be invoked.  An execution context may not be
     * started until the RT-Components that participate in it have
     * been initialized.  An execution context may be started and
     * stopped multiple times.
     *
     * @return The return code of ReturnCode_t type
     *
     * @endif
     */
    RTC::ReturnCode_t start(void);

    /*!
     * @if jp
     * @brief ExecutionContext ̎s~
     *
     * ExecutionContext ̏Ԃ Stopped Ƃ邽߂̃NGXg𔭍s
     * BJڂꍇ́AComponentAction::on_shutdown Ăяo
     * BQĂRTR|[lgIO
     * ExecutionContext ~KvBExecutionContext ͕
     * Jn/~JԂƂłB
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Stop the ExecutionContext
     *
     * Request that the context enter the Stopped state.  Once the
     * transition occurs, the ComponentAction::on_shutdown operation
     * will be invoked.  An execution context must be stopped before
     * the RT components that participate in it are finalized.  An
     * execution context may be started and stopped multiple times.
     *
     * @return The return code of ReturnCode_t type
     *
     * @endif
     */
    RTC::ReturnCode_t stop(void);

    RTC::ReturnCode_t startThread(void);

    RTC::ReturnCode_t stopThread(void);

    
    /*!
     * @if jp
     * @brief ExecutionContext ̎sύXꂽ
     *
     * ExecutionContext ̎sύXꂽꍇ̃R[obNĂ
     * oBEC̎NX set_rate() Iy[VĂяoꂽ
     * AExecutionContextProfile::setRate() ĂяoAvt@C
     * ̎lXVƂƂɁÅ֐Ă΂A
     * RTObjectStateMachine::onRateChagned() ĂяoB
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Execution rate of ExecutionContext chnaged
     *
     * A call back for execution rate chang when ExecutionContext's
     * execution rate. If set_rate() operation is called,
     * ExecutionContextProfile::setRate() will be called, execution
     * rate in the profile will be updated, and
     * RTObjectStateMachine::onRateChagned() will be called with this
     * member function.
     *
     * @return The return code of ReturnCode_t type
     *
     * @endif
     */
    RTC::ReturnCode_t rateChanged();

    /*!
     * @if jp
     * @brief RTR|[lgANeBu
     *
     * Inactive ԂɂRTR|[lgActive ɑJڂAANeB
     * uB̑삪Ă΂ꂽʁAon_activate ĂяoBw
     * 肵RTR|[lgQ҃XgɊ܂܂Ȃꍇ́A
     * BAD_PARAMETER ԂBw肵RTR|[lg̏Ԃ
     * Inactive ȊȌꍇ́APRECONDITION_NOT_MET ԂB
     *
     * @param comp ANeBuΏRTR|[lg
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Activate an RT-component
     *
     * The given participant RTC is Inactive and is therefore not
     * being invoked according to the execution contextfs execution
     * kind. This operation shall cause the RTC to transition to the
     * Active state such that it may subsequently be invoked in this
     * execution context.  The callback on_activate shall be called as
     * a result of calling this operation. This operation shall not
     * return until the callback has returned, and shall result in an
     * error if the callback does.
     *
     * @param comp The target RT-Component for activation
     *
     * @return The return code of ReturnCode_t type
     *
     * @endif
     */
    RTC::ReturnCode_t activateComponent(RTC::LightweightRTObject_ptr comp,
                                        RTObjectStateMachine*& rtobj);
    RTC::ReturnCode_t waitActivateComplete(RTObjectStateMachine*& rtobj,
                                           coil::TimeValue timeout = 1.0,
                                           long int cycle = 1000);
    /*!
     * @if jp
     * @brief RTR|[lgANeBu
     *
     * Inactive ԂɂRTR|[lgANeBuAInactive
     * ɑJڂB̑삪Ă΂ꂽʁAon_deactivate Ăяo
     * Bw肵RTR|[lgQ҃XgɊ܂܂Ȃꍇ́A
     * BAD_PARAMETER ԂBw肵RTR|[lg̏Ԃ
     * Active ȊȌꍇ́APRECONDITION_NOT_MET ԂB
     *a
     * @param comp ANeBuΏRTR|[lg
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Deactivate an RT-component
     *
     * The given RTC is Active in the execution context. Cause it to
     * transition to the Inactive state such that it will not be
     * subsequently invoked from the context unless and until it is
     * activated again.  The callback on_deactivate shall be called as
     * a result of calling this operation. This operation shall not
     * return until the callback has returned, and shall result in an
     * error if the callback does.
     *
     * @param comp The target RT-Component for deactivate
     *
     * @return The return code of ReturnCode_t type
     *
     * @endif
     */
    RTC::ReturnCode_t deactivateComponent(RTC::LightweightRTObject_ptr comp,
                                          RTObjectStateMachine*& rtobj);
    RTC::ReturnCode_t waitDeactivateComplete(RTObjectStateMachine*& rtobj,
                                             coil::TimeValue timeout = 1.0,
                                             long int cycle = 1000);

    /*!
     * @if jp
     * @brief RTR|[lgZbg
     *
     * Error ԂRTR|[lg̕A݂B̑삪Ă΂ꂽ
     * ʁAon_reset ĂяoBw肵RTR|[lgQ҃
     * XgɊ܂܂Ȃꍇ́ABAD_PARAMETER ԂBw肵RTR
     * |[lg̏Ԃ Error ȊȌꍇ́APRECONDITION_NOT_MET 
     * B
     *
     * @param comp ZbgΏRTR|[lg
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Reset the RT-component
     *
     * Attempt to recover the RTC when it is in Error.  The
     * ComponentAction::on_reset callback shall be invoked. This
     * operation shall not return until the callback has returned, and
     * shall result in an error if the callback does. If possible, the
     * RTC developer should implement that callback such that the RTC
     * may be returned to a valid state.
     *
     * @param comp The target RT-Component for reset
     *
     * @return The return code of ReturnCode_t type
     *
     * @endif
     */
    RTC::ReturnCode_t resetComponent(RTC::LightweightRTObject_ptr com,
                                     RTObjectStateMachine*& rtobj);
    RTC::ReturnCode_t waitResetComplete(RTObjectStateMachine*& rtobj,
                                        coil::TimeValue timeout = 1.0,
                                        long int cycle = 1000);

    /*!
     * @if jp
     * @brief RTR|[lg̏Ԃ擾
     *
     * w肵RTR|[lg̏(LifeCycleState)擾Bw肵
     * RTR|[lgQ҃XgɊ܂܂Ȃꍇ́A
     * UNKNOWN_STATE ԂB
     *
     * @param comp Ԏ擾ΏRTR|[lg
     *
     * @return ݂̏(LifeCycleState)
     *
     * @else
     *
     * @brief Get RT-component's state
     *
     * This operation shall report the LifeCycleState of the given
     * participant RTC.  UNKNOWN_STATE will be returned, if the given
     * RT-Component is not inclued in the participant list.
     *
     * @param comp The target RT-Component to get the state
     *
     * @return The current state of the target RT-Component(LifeCycleState)
     *
     * @endif
     */
    RTC::LifeCycleState getComponentState(RTC::LightweightRTObject_ptr comp);
    const char* getStateString(RTC::LifeCycleState state)
    {
      const char* st[] = {
        "CREATED_STATE",
        "INACTIVE_STATE",
        "ACTIVE_STATE",
        "ERROR_STATE"
      };
      return state >= RTC::CREATED_STATE && state <= RTC::ERROR_STATE ?
        st[state] : "";
    }
    /*!
     * @if jp
     * @brief RTR|[lgǉ
     *
     * w肵RTR|[lg"Q҃Xg"ɒǉBǉꂽ
     * RTR|[lg attach_context Ă΂AInactive ԂɑJ
     * Bw肳ꂽRTR|[lg null ̏ꍇ́ABAD_PARAMETER 
     * ԂBw肳ꂽRTR|[lg DataFlowComponent ȊO
     * ꍇ́ABAD_PARAMETER ԂB
     *
     * ̊֐Ă΂ꂽiKł́A RTC  StateMachine ̃R[
     * obN֐Ăяỏ\A܂ "Q҃Xg:
     * m_addedComps"  RTC ǉĂBR[obNĂяoU
     * IiKœI updateComponentList() ֐Ă΂A
     * Worker ɐ RTC QԂƂȂB
     *
     * @param comp ǉΏRTR|[lg
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Add an RT-component
     *
     * The operation causes the given RTC to begin participating in
     * the execution context.  The newly added RTC will receive a call
     * to LightweightRTComponent::attach_context and then enter the
     * Inactive state.  BAD_PARAMETER will be invoked, if the given
     * RT-Component is null or if the given RT-Component is other than
     * DataFlowComponent.
     *
     * When right after this operation is called, since other RTCs
     * might being called their callbacks of the StateMachine, at
     * first the RTC is added into the "Participants: m_addedComps."
     * After RTCs callbacks are called, updateComponentList() function
     * is invoked and RTCs are officially a member of the worker.
     *
     * @param comp The target RT-Component for add
     *
     * @return The return code of ReturnCode_t type
     *
     * @endif
     */
    RTC::ReturnCode_t addComponent(RTC::LightweightRTObject_ptr comp);

    /*!
     * @if jp
     * @brief R|[lgoChB
     *
     * R|[lg owned context oChBɓn
     * RTObject_impl ́A EC  owner łAIɂ RTObject
     *  bindContext() R[Aec_id擾Bec_id 
     * ECOTHER_OFFSET B̌ARTObject ̃vLVł
     * RTObjectStateMachine 𐶐AR|[lg̃XgɉB
     *
     * @param rtc RTR|[lg
     * @return ReturnCode_t ^̃^[R[h
     * @else
     * @brief Bind the component.
     *
     * This operation bind a RT-Component into owned context. Given
     * argument RTObject_impl is the owner of the EC, and internally
     * RTObject's bindContext() member function is called and obtain
     * ec_id. The ec_id must be smaller than ECOTHER_OFFSET. After
     * that proxy of RTObject RTObjectStateMachine is created, and it
     * is added to the component list.
     *
     * @param rtc RT-Component's instances
     * @return The return code of ReturnCode_t type
     * @endif
     */
    RTC::ReturnCode_t bindComponent(RTC::RTObject_impl* rtc);

    /*!
     * @if jp
     * @brief RTR|[lgQ҃Xg폜
     *
     * w肵RTR|[lgQ҃Xg폜B폜ꂽ
     * RTR|[lg detach_context Ă΂Bw肳ꂽRTR|[
     * lgQ҃Xgɓo^ĂȂꍇ́ABAD_PARAMETER 
     * B
     *
     * @param comp 폜ΏRTR|[lg
     *
     * @return ReturnCode_t ^̃^[R[h
     *
     * @else
     *
     * @brief Remove the RT-Component from participant list
     *
     * This operation causes a participant RTC to stop participating in the
     * execution context.
     * The removed RTC will receive a call to
     * LightweightRTComponent::detach_context.
     * BAD_PARAMETER will be returned, if the given RT-Component is not 
     * participating in the participant list.
     *
     * @param comp The target RT-Component for delete
     *
     * @return The return code of ReturnCode_t type
     *
     * @endif
     */
    RTC::ReturnCode_t removeComponent(RTC::LightweightRTObject_ptr comp);
    RTObjectStateMachine* findComponent(RTC::LightweightRTObject_ptr comp);

    bool isAllCurrentState(RTC::LifeCycleState state);
    bool isAllNextState(RTC::LifeCycleState state);
    bool isOneOfCurrentState(RTC::LifeCycleState state);
    bool isOneOfNextState(RTC::LifeCycleState state);

    void invokeWorker();
    void invokeWorkerPreDo();
    void invokeWorkerDo();
    void invokeWorkerPostDo();

    /*!
     * @if jp
     * @brief R|[lgXg̍XV
     * @else
     * @brief Updating component list
     * @endif
     */
    void updateComponentList();

    //------------------------------------------------------------
    // member variables
  protected:
    /*!
     * @if jp
     * @brief K[Xg[
     * @else
     * @brief Logger stream
     * @endif
     */
    RTC::Logger rtclog;

    RTC::ExecutionContextService_var m_ref;

    /*!
     * @if jp
     * @brief ExecutionContext ̎s
     * true: running, false: stopped
     * @else
     * @brief The running state of ExecutionContext
     * true: running, false: stopped
     * @endif
     */
    bool m_running;

    /*!
     * @if jp
     * @brief R|[lg̎Q҃Xg
     * @else
     * @brief List of the participating component
     * @endif
     */
    std::vector<RTC_impl::RTObjectStateMachine*> m_comps;
    mutable coil::Mutex m_mutex;
    std::vector<RTC_impl::RTObjectStateMachine*> m_addedComps;
    mutable coil::Mutex m_addedMutex;
    std::vector<RTC_impl::RTObjectStateMachine*> m_removedComps;
    mutable coil::Mutex m_removedMutex;
    typedef std::vector<RTC_impl::RTObjectStateMachine*>::iterator CompItr;

  }; // class PeriodicExecutionContext
}; // namespace RTC

#ifdef WIN32
#pragma warning( default : 4290 )
#endif

#endif // RTC_PERIODICEXECUTIONCONTEXT_H
