// -*- C++ -*-
/*!
 * @file  InPortCorbaCdrConsumer.h
 * @brief InPortCorbaCdrConsumer class
 * @date  $Date: 2007-12-31 03:08:03 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: InPortCorbaCdrConsumer.h 1255 2009-04-07 01:09:47Z kurihara $
 *
 */

#ifndef RTC_INPORTCORBACDRCONSUMER_H
#define RTC_INPORTCORBACDRCONSUMER_H


//#include <rtm/BufferBase.h>

#include <rtm/idl/DataPortSkel.h>
#include <rtm/CorbaConsumer.h>
#include <rtm/InPortConsumer.h>
#include <rtm/Manager.h>

namespace RTC
{
  /*!
   * @if jp
   * @class InPortCorbaCdrConsumer
   * @brief InPortCorbaCdrConsumer NX
   *
   * InPortConsumer 
   *
   * f[^] CORBA  OpenRTM::InPortCdr C^[tF[X𗘗p
   * Apush ^f[^t[^ InPort RV[}NXB
   *
   * @since 0.4.0
   *
   * @else
   * @class InPortCorbaCdrConsumer
   * @brief InPortCorbaCdrConsumer class
   *
   * The InPort consumer class which uses the OpenRTM::InPortCdr
   * interface in CORBA for data transfer and realizes a push-type
   * dataflow.
   *
   * @since 0.4.0
   *
   * @endif
   */
  class InPortCorbaCdrConsumer
    : public InPortConsumer,
      public CorbaConsumer< ::OpenRTM::InPortCdr >
  {
  public:
    DATAPORTSTATUS_ENUM
    /*!
     * @if jp
     * @brief RXgN^
     *
     * RXgN^
     *
     * @param buffer YRV[}Ɋ蓖Ăobt@IuWFNg
     *
     * @else
     * @brief Constructor
     *
     * Constructor
     *
     * @param buffer The buffer object that is attached to this Consumer
     *
     * @endif
     */
    InPortCorbaCdrConsumer(void);
    
    /*!
     * @if jp
     * @brief fXgN^
     *
     * fXgN^
     *
     * @else
     * @brief Destructor
     *
     * Destructor
     *
     * @endif
     */
    virtual ~InPortCorbaCdrConsumer(void);

    /*!
     * @if jp
     * @brief ݒ菉
     *
     * InPortConsumer̊eݒsBNXł́A^ꂽ
     * PropertiesKvȏ擾ĊeݒsB init() 
     * ́AInPortProviderエсAڑɂꂼĂ΂
     * \BāÅ֐͕Ă΂邱Ƃz肵ċL
     * qׂłB
     * 
     * @param prop ݒ
     *
     * @else
     *
     * @brief Initializing configuration
     *
     * This operation would be called to configure in initialization.
     * In the concrete class, configuration should be performed
     * getting appropriate information from the given Properties data.
     * This function might be called right after instantiation and
     * connection sequence respectivly.  Therefore, this function
     * should be implemented assuming multiple call.
     *
     * @param prop Configuration information
     *
     * @endif
     */
    virtual void init(coil::Properties& prop);

    /*!
     * @if jp
     * @brief ڑւ̃f[^M
     *
     * ڑ̃|[gփf[^𑗐M邽߂̏z֐B
     * 
     * ̊֐́Aȉ̃^[R[hԂB
     *
     * - PORT_OK:       IB
     * - PORT_ERROR:    f[^M̉ߒŉ炩̃G[B
     * - SEND_FULL:     f[^𑗐MA葤obt@tB
     * - SEND_TIMEOUT:  f[^𑗐MA葤obt@^CAEgB
     * - UNKNOWN_ERROR: s̃G[
     *
     * @param data Mf[^
     * @return ^[R[h
     *
     * @else
     * @brief Send data to the destination port
     *
     * Pure virtual function to send data to the destination port.
     *
     * This function might the following return codes
     *
     * - PORT_OK:       Normal return
     * - PORT_ERROR:    Error occurred in data transfer process
     * - SEND_FULL:     Buffer full although OutPort tried to send data
     * - SEND_TIMEOUT:  Timeout although OutPort tried to send data
     * - UNKNOWN_ERROR: Unknown error
     *
     * @endif
     */
    virtual ReturnCode put(cdrMemoryStream& data);

    /*!
     * @if jp
     * @brief InterfaceProfileJ
     *
     * InterfaceProfileJB
     * Ŏw肷vpeB NameValue IuWFNg
     * dataport.interface_type l𒲂ׁAY|[gɐݒ肳Ă
     * C^[tF[X^Cvƈvꍇ̂ݏ擾B
     *
     * @param properties InterfaceProfile󂯎vpeB
     *
     * @else
     * @brief Publish InterfaceProfile information
     *
     * Publish interfaceProfile information.
     * Check the dataport.interface_type value of the NameValue object 
     * specified by an argument in property information and get information
     * only when the interface type of the specified port is matched.
     *
     * @param properties Properties to get InterfaceProfile information
     *
     * @endif
     */
    virtual void publishInterfaceProfile(SDOPackage::NVList& properties);

    /*!
     * @if jp
     * @brief f[^Mʒmւ̓o^
     *
     * w肳ꂽvpeBɊÂāAf[^oʒm̎󂯎ɓo^B
     *
     * @param properties o^
     *
     * @return o^(o^:trueAo^s:false)
     *
     * @else
     * @brief Subscribe to the data sending notification
     *
     * Subscribe to the data sending notification based on specified 
     * property information.
     *
     * @param properties Information for subscription
     *
     * @return Subscription result (Successful:true, Failed:false)
     *
     * @endif
     */
    virtual bool subscribeInterface(const SDOPackage::NVList& properties);
    
    /*!
     * @if jp
     * @brief f[^Mʒm̓o^
     *
     * f[^oʒm̎󂯎肩o^B
     *
     * @param properties o^
     *
     * @else
     * @brief Unsubscribe the data send notification
     *
     * Unsubscribe the data send notification.
     *
     * @param properties Information for unsubscription
     *
     * @endif
     */
    virtual void unsubscribeInterface(const SDOPackage::NVList& properties);

  private:
    /*!
     * @if jp
     * @brief IOR񂩂IuWFNgQƂ擾
     *
     * @return true: 擾, false: 擾s
     *
     * @else
     * @brief Getting object reference fromn IOR string
     *
     * @return true: succeeded, false: failed
     *
     * @endif
     */
    bool subscribeFromIor(const SDOPackage::NVList& properties);

    /*!
     * @if jp
     * @brief Any璼ڃIuWFNgQƂ擾
     *
     * @return true: 擾, false: 擾s
     *
     * @else
     * @brief Getting object reference fromn Any directry
     *
     * @return true: succeeded, false: failed
     *
     * @endif
     */
    bool subscribeFromRef(const SDOPackage::NVList& properties);

    /*!
     * @if jp
     * @brief ڑ(IOR)
     *
     * @return true: 擾, false: 擾s
     *
     * @else
     * @brief ubsubscribing (IOR version)
     *
     * @return true: succeeded, false: failed
     *
     * @endif
     */
    bool unsubscribeFromIor(const SDOPackage::NVList& properties);

    /*!
     * @if jp
     * @brief ڑ(Object reference)
     *
     * @return true: 擾, false: 擾s
     *
     * @else
     * @brief ubsubscribing (Object reference version)
     *
     * @return true: succeeded, false: failed
     *
     * @endif
     */
    bool unsubscribeFromRef(const SDOPackage::NVList& properties);
    
  private:
    /*!
     * @if jp
     * @brief ^[R[hϊ
     * @else
     * @brief Return codes conversion
     * @endif
     */
    InPortConsumer::ReturnCode convertReturnCode(OpenRTM::PortStatus ret);

    mutable Logger rtclog;
    coil::Properties m_properties;
  };
};     // namespace RTC

extern "C"
{
  /*!
   * @if jp
   * @brief W[֐
   *
   * InPortCorbaCdrConsumer ̃t@Ngo^鏉֐B
   *
   * @else
   * @brief Module initialization
   *
   * This initialization function registers InPortCorbaCdrConsumer's factory.
   *
   * @endif
   */
  void InPortCorbaCdrConsumerInit(void);
};

#endif // RTC_INPORTCORBACDRCONSUMER_H

